// =============================================================== //
//                                                                 //
//   File      : db_scanner.cxx                                    //
//   Purpose   :                                                   //
//                                                                 //
//   Institute of Microbiology (Technical University Munich)       //
//   http://www.arb-home.de/                                       //
//                                                                 //
// =============================================================== //

#include <db_scanner.hxx>

#include <AW_rename.hxx>

#include <aw_select.hxx>
#include <aw_msg.hxx>
#include <arb_progress.h>
#include <aw_root.hxx>

#include <arb_str.h>
#include <arb_strbuf.h>
#include <ad_cb.h>

#include <map>

void DbScanner::editfield_value_changed_cb() {
    if (ignore_editfield_change) return;

    LocallyModify<bool> do_not_recurse(ignore_editfield_change, true);

    bool update_self = false;
    {
        GB_begin_transaction(gb_main);

        GB_ERROR    error      = NULp;
        const char *field_name = awar_selected_field->read_char_pntr();

        if (!gb_item) {
            error = "No item selected";
        }
        else if (!field_name[0]) {
            error = "No field selected";
        }
        else if (!awar_edit_enabled->read_int()) { // edit disabled
            awar_editfield->write_string("");
            error = "Change ignored (edit is disabled!)";
        }
        else {
            char *value    = awar_editfield->read_string();
            int   vlen     = strlen(value);

            while (vlen>0 && value[vlen-1] == '\n') vlen--; // remove trailing newlines
            value[vlen]     = 0;

            GBDATA *gbd = search_selected_field();

            if (!gbd) { // field does not exist -> create new element
                if (vlen) {
                    GBDATA *gb_key      = GBT_get_changekey(gb_main, field_name, selector.change_key_path);
                    GBDATA *gb_key_type = GB_entry(gb_key, CHANGEKEY_TYPE);
                    GBDATA *gb_new      = GB_search(gb_item, field_name, (GB_TYPES)GB_read_int(gb_key_type));
                    if (!gb_new) error  = GB_await_error();
                    else    error       = GB_write_autoconv_string(gb_new, value);
                }
            }
            else { // change old element
                if (strcmp(field_name, "name") == 0) { // This is a real rename !!! // @@@ should use MutableItemSelector::id_field (no real problem)
                    if (selector.type == QUERY_ITEM_SPECIES) { // species
                        arb_progress  progress("Manual change of species ID");
                        char         *name = nulldup(GBT_read_name(gb_item));

                        if (strlen(value)) {
                            GBT_begin_rename_session(gb_main, 0);

                            error = GBT_rename_species(name, value, false);

                            if (error) GBT_abort_rename_session();
                            else error = GBT_commit_rename_session();
                        }
                        else {
                            error = AWTC_recreate_name(gb_item);
                        }

                        free(name);
                    }
                    else { // non-species (gene, experiment, etc.)
                        if (strlen(value)) {
                            GBDATA *gb_exists    = NULp;
                            GBDATA *gb_item_data = GB_get_father(gb_item);

                            for (gb_exists = selector.get_first_item(gb_item_data, QUERY_ALL_ITEMS);
                                 gb_exists;
                                 gb_exists = selector.get_next_item(gb_exists, QUERY_ALL_ITEMS))
                            {
                                if (ARB_stricmp(GBT_read_name(gb_exists), value) == 0) break;
                            }

                            if (gb_exists) error = GBS_global_string("There is already a %s named '%s'", selector.item_name, value);
                            else error           = GB_write_autoconv_string(gbd, value);
                        }
                        else {
                            error = "The 'name' field can't be empty.";
                        }
                    }

                    if (!error) update_self = true;
                }
                else {
                    if (vlen) {
                        error = GB_write_autoconv_string(gbd, value);
                    }
                    else {
                        if (GB_child(gbd)) {
                            error = "Sorry, cannot perform a deletion.\n(The selected entry has child entries. Delete them first.)";
                        }
                        else {
                            error = GB_delete(gbd);
                        }
                    }
                }
            }
            free(value);
        }

        aw_message_if(GB_end_transaction(gb_main, error));
    }

    awar_selected_field->touch(); // force refresh of edit-box

    if (update_self) { // if the name changed -> rewrite awars AFTER transaction was closed
        GB_transaction ta(gb_main);

        char *my_id = selector.generate_item_id(gb_main, gb_item);
        selector.update_item_awars(gb_main, get_root(), my_id); // update awars (e.g. AWAR_SPECIES_NAME)
        free(my_id);
    }
}

void DbScanner::toggle_marked_cb() {
    if (gb_item && !ignore_marktoggle_change) {
        long flag = awar_mark->read_int();
        GB_transaction ta(gb_main);
        GB_write_flag(gb_item, flag);
    }
}

void DbScanner::remap_edit_box() {
    GB_transaction ta(gb_main);

    GBDATA *gb_wanted_field = NULp;
    if (awar_edit_enabled->read_int()) {
        gb_wanted_field = search_selected_field(); // map only if editing is allowed
    }

    if (gb_wanted_field != gb_field) {
        DatabaseCallback editbox_update_cb = makeDatabaseCallback(DbScanner::field_changed_cb, this);
        if (gb_field) GB_remove_callback(gb_field, GB_CB_CHANGED_OR_DELETED, editbox_update_cb);

        gb_field = gb_wanted_field; // change field

        if (gb_field) {
            GB_add_callback(gb_field, GB_CB_CHANGED_OR_DELETED, editbox_update_cb);
        }
        field_changed_cb(GB_CB_CHANGED);
    }
}

inline const char *awarname(const char *scanner_id, const char *entry) {
    return GBS_global_string("tmp/dbscan/%s/%s", scanner_id, entry);
}

void DbScanner::create_awars(const char *scanner_id, bool have_edit_field, bool have_edit_toggle, bool have_mark_toggle) {
    arb_assert(implicated(have_edit_toggle, have_edit_field));

    // create scanner-local AWARS
    const char *list_awarname = awarname(scanner_id, "list");
    if (get_root()->awar_no_error(list_awarname)) { // awar already exists
        GBK_terminatef("multiple scanners bound to AWAR '%s'", list_awarname);
    }
    awar_selected_field = get_root()->awar_string(list_awarname, "", AW_ROOT_DEFAULT);

    awar_mapped_item_ID = get_root()->awar_string(awarname(scanner_id, "id"), "<undef>", AW_ROOT_DEFAULT);
    if (have_edit_field) {
        awar_editfield = get_root()->awar_string(awarname(scanner_id, "edit"), "", AW_ROOT_DEFAULT);
        if (have_edit_toggle) {
            awar_edit_enabled = get_root()->awar_int(awarname(scanner_id, "edit_enable"), true, AW_ROOT_DEFAULT);
        }
    }
    if (have_mark_toggle) {
        awar_mark = get_root()->awar_int(awarname(scanner_id, "mark"), true, AW_ROOT_DEFAULT);
    }
}

void DbScanner::create_field_edit_widgets(const char *edit_pos_fig, const char *edit_enable_pos_fig) {
    arb_assert(edit_pos_fig);

    RootCallback   remap_rcb = makeRootCallback(DbScanner::remap_edit_box, this);
    WindowCallback remap_wcb = makeWindowCallback(DbScanner::remap_edit_box, this);

    if (edit_enable_pos_fig) {
        aws->at(edit_enable_pos_fig);
        aws->create_toggle(awar_edit_enabled->awar_name);
    }

    awar_selected_field->add_callback(remap_rcb);
    awar_mapped_item_ID->add_callback(remap_rcb); // refresh edit-field if item changes
    if (edit_enable_pos_fig) awar_edit_enabled->add_callback(remap_rcb);
    awar_editfield->add_callback(makeRootCallback(DbScanner::editfield_value_changed_cb, this));

    aws->at(edit_pos_fig);
    aws->create_text_field(awar_editfield->awar_name, 20, 10);
}


DbScanner *DbScanner::create(GBDATA         *gb_main,
                             const char     *scanner_id,
                             AW_window      *aws,
                             const char     *box_pos_fig,
                             const char     *edit_pos_fig,
                             const char     *edit_enable_pos_fig,
                             DB_SCANNERMODE  scannermode,
                             const char     *mark_pos_fig,
                             ItemSelector&   selector)
{
    /* create an unmapped scanner box and optionally some buttons,
     * the return value is used as handle to further scanner functions
     */

    DbScanner *scanner = new DbScanner(scannermode, selector, aws, gb_main);

    if (gb_main) GB_push_transaction(gb_main);

    scanner->create_awars(scanner_id, edit_pos_fig, edit_enable_pos_fig, mark_pos_fig);

    // create GUI elements
    aws->at(box_pos_fig);
    scanner->create_field_selection_list();
    if (mark_pos_fig) {
        aws->at(mark_pos_fig);
        scanner->create_mark_toggle();
    }
    if (edit_pos_fig) {
        scanner->create_field_edit_widgets(edit_pos_fig, edit_enable_pos_fig);
    }

    if (gb_main) GB_pop_transaction(gb_main);
    aws->set_popup_callback(makeWindowCallback(DbScanner::changed_cb, scanner, GB_CB_CHANGED));
    return scanner;
}

typedef std::map<const char *, int> KeyOccur;

inline char *make_unique_key(const char *key, KeyOccur& occurred) {
    KeyOccur::iterator found = occurred.find(key);
    int                count = 1;
    if (found == occurred.end()) {
        occurred[key] = count;
    }
    else {
        count = ++found->second;
    }
    return GBS_global_string_copy("%s_%i", key, count);
}

void DbScanner::scan_fields_recursive(GBDATA *gbd, const int depth) {
    static KeyOccur       occurred;
    static GBS_strstruct *out = NULp;

    // init static variables (at top of recursion):
    if (!depth) {
        out = new GBS_strstruct(1000);
    }

    arb_assert(out->empty());
    for (int i = 0; i<depth; ++i) out->ncat(": ", 2);

    const char *key = GB_read_key_pntr(gbd);
    out->nprintf(64+2, "%-14s", key); // see GB_KEY_LEN_MAX

    char *id = make_unique_key(key, occurred); // create unique IDs for fields with same key
    if (GB_read_type(gbd) == GB_DB) {
        out->ncat("<CONTAINER>:", 12);
        field_sel->insert(out->get_data(), id);
        out->erase();

        for (GBDATA *gb2 = GB_child(gbd); gb2; gb2 = GB_nextChild(gb2)) {
            scan_fields_recursive(gb2, depth+1);
        }
    }
    else {
        char *data = GB_read_as_string(gbd);
        if (data) {
            out->cat(data);
            free(data);
        }
        else {
            out->ncat("<unprintable>", 13);
        }
        field_sel->insert(out->get_data(), id);
        out->erase();
    }
    free(id);

    // cleanup static variables (at top of recursion):
    if (!depth) {
        occurred.clear();
        delete out; out = NULp;
    }
}

void DbScanner::scan_list() {
    const int INFO_WIDTH = 1000;

  refresh_again :
    char buffer[INFO_WIDTH+1]; // @@@ as well use GBS_strstruct here like in scan_fields_recursive?
    memset(buffer, 0, INFO_WIDTH+1);

    static int last_max_name_width;
    int        max_name_width = 0;

    if (last_max_name_width == 0) last_max_name_width = 15;

    field_sel->clear();

    GBDATA *gb_key_data = GB_search(gb_main, get_selector().change_key_path, GB_CREATE_CONTAINER);

    for (int existing = 1; existing >= 0; --existing) {
        for (GBDATA *gb_key = GB_entry(gb_key_data, CHANGEKEY); gb_key; gb_key = GB_nextEntry(gb_key)) {
            GBDATA *gb_key_hidden = GB_entry(gb_key, CHANGEKEY_HIDDEN);
            if (gb_key_hidden && GB_read_int(gb_key_hidden)) continue; // don't show hidden fields in 'species information' window

            GBDATA *gb_key_name = GB_entry(gb_key, CHANGEKEY_NAME);
            if (!gb_key_name) continue;

            GBDATA *gb_key_type = GB_entry(gb_key, CHANGEKEY_TYPE);

            const char *name = GB_read_char_pntr(gb_key_name);
            GBDATA     *gbd  = GB_search(gb_item, name, GB_FIND);

            if ((!existing) == (!gbd)) { // first print only existing; then non-existing entries
                char *p      = buffer;
                int   len    = sprintf(p, "%-*s %c", last_max_name_width, name, GB_type_2_char((GB_TYPES)GB_read_int(gb_key_type)));

                p += len;

                int name_width = strlen(name);
                if (name_width>max_name_width) max_name_width = name_width;

                if (gbd) {      // existing entry
                    *(p++) = GB_read_security_write(gbd)+'0';
                    *(p++) = ':';
                    *(p++) = ' ';
                    *p     = 0;

                    {
                        char *data = GB_read_as_string(gbd);
                        if (data) {
                            int rest  = INFO_WIDTH-(p-buffer);
                            int ssize = strlen(data);

                            if (ssize > rest) {
                                ssize = GBS_shorten_repeated_data(data);
                                if (ssize > rest) {
                                    if (ssize>5) strcpy(data+rest-5, "[...]");
                                    ssize = rest;
                                }
                            }

                            memcpy(p, data, ssize);
                            p[ssize] = 0;

                            free(data);
                        }
                    }
                    field_sel->insert(buffer, name);
                }
                else { // non-existing entry
                    p[0] = ' '; p[1] = ':'; p[2] = 0;
                    field_sel->insert(buffer, name);
                }
            }
        }
    }

    if (last_max_name_width < max_name_width) {
        last_max_name_width = max_name_width;
        goto refresh_again;
    }
}

void DbScanner::field_changed_cb(GB_CB_TYPE cbtype) {
    if (cbtype == GB_CB_DELETE) {
        gb_field = NULp;
    }

    LocallyModify<bool> just_a_refresh(ignore_editfield_change, true);

    if (gb_field) {
        char *data = GB_read_as_string(gb_field);
        if (!data) data = strdup("<you cannot edit this datatype>");
        awar_editfield->write_string(data);
        free(data);
    }
    else {
        awar_editfield->write_string("");
    }
}

void DbScanner::changed_cb(GB_CB_TYPE cbtype) {
    // performs updates necessary when mapped item changes:
    // - refresh field selection list
    // - update display of item name

    if (cbtype == GB_CB_DELETE) {
        gb_item = NULp;
    }
    if (gb_item && !aws->is_shown()) {
        return;
    }

    // update selection list
    field_sel->clear();
    if (gb_item) {
        GB_transaction ta(gb_main);

        switch (scannermode) {
            case DB_SCANNER: scan_fields_recursive(gb_item, 0); break;
            case DB_VIEWER:  scan_list(); break;
        }
    }

    field_sel->insert_default("", "");
    field_sel->update();

    if (gb_item) {
        GB_transaction ta(gb_main);

        if (awar_mark) {
            LocallyModify<bool> ignoreUpdate(ignore_marktoggle_change, true);
            long flag = GB_read_flag(gb_item);
            awar_mark->write_int(flag);
        }

        char *id = get_mapped_item_id();
        awar_mapped_item_ID->write_string(id);
        free(id);
    }
    else {
        awar_mapped_item_ID->write_string("<none selected>");
    }
    awar_selected_field->touch();
}

void DbScanner::keydata_modified_cb(GB_CB_TYPE cbtype) {
    // unmap edit field if 'key_data' has been changed (entries might have been deleted)

    bool deselect = false;
    {
        const char *field_name = awar_selected_field->read_char_pntr();
        if (field_name[0]) {
            GBDATA *gb_key = GBT_get_changekey(gb_main, field_name, selector.change_key_path);

            if (!gb_key) { // key unknown
                deselect = true;
            }
            else {
                GBDATA *gb_key_hidden = GB_entry(gb_key, CHANGEKEY_HIDDEN);
                if (gb_key_hidden && GB_read_int(gb_key_hidden)) { // key hidden -> deselect
                    deselect = true;
                }
            }
        }
    }

    if (deselect) awar_selected_field->write_string(""); // deselect field
    changed_cb(cbtype);
}

void DbScanner::Map(GBDATA *gb_new_item, const char *key_path) {
    GB_transaction ta(gb_main);

    GBDATA *gb_key_data = GB_search(gb_main, key_path, GB_CREATE_CONTAINER);

    if (gb_item) {
        GB_remove_callback(gb_item, GB_CB_CHANGED_OR_DELETED, makeDatabaseCallback(DbScanner::changed_cb, this));
        if (scannermode == DB_VIEWER) {
            GB_remove_callback(gb_key_data, GB_CB_CHANGED, makeDatabaseCallback(DbScanner::keydata_modified_cb, this));
        }
    }

    gb_item = gb_new_item;

    if (gb_item) {
        GB_add_callback(gb_item, GB_CB_CHANGED_OR_DELETED, makeDatabaseCallback(DbScanner::changed_cb, this));
        if (scannermode == DB_VIEWER) {
            GB_add_callback(gb_key_data, GB_CB_CHANGED, makeDatabaseCallback(DbScanner::keydata_modified_cb, this));
        }
    }

    changed_cb(GB_CB_CHANGED);
}
