// ==================================================================== //
//                                                                      //
//   File      : db_scanner.hxx                                         //
//   Purpose   : ARB database scanner                                   //
//                                                                      //
//                                                                      //
// Coded by Ralf Westram (coder@reallysoft.de) in May 2005              //
// Copyright Department of Microbiology (Technical University Munich)   //
//                                                                      //
// Visit our web site at: http://www.arb-home.de/                       //
//                                                                      //
// ==================================================================== //
#ifndef DB_SCANNER_HXX
#define DB_SCANNER_HXX

#ifndef ARBDB_H
#include <arbdb.h>
#endif
#ifndef AW_WINDOW_HXX
#include <aw_window.hxx>
#endif
#ifndef AW_AWAR_HXX
#include <aw_awar.hxx>
#endif
#ifndef ITEMS_H
#include <items.h>
#endif

enum DB_SCANNERMODE {
    DB_SCANNER, // used for SAI-info and for SAI/species-info in merge-tool       (recursively show all existing fields; order as occurring in DB)
    DB_VIEWER   // used for item-info (species, gene, experiment)                 (show all known fields; used at top, unused at bottom; user-defined order)
};

class DbScanner { // @@@ derive from AW_DB_selection?
    // Basically displays DB-fields of a selected item
    // (e.g. a species or an SAI) in a field-selection list.
    //
    // - optionally provides an edit-box (allowing to edit the selected field)
    // - optionally provides a mark-toggle (show/edit mark-flag of item)

    ItemSelector&  selector;
    AW_window     *aws;

    GBDATA *gb_main;

    AW_selection_list *field_sel;

    GBDATA *gb_item;  // currently mapped item (e.g. species)
    GBDATA *gb_field; // currently selected field

    DB_SCANNERMODE scannermode;

    bool ignore_editfield_change;
    bool ignore_marktoggle_change;

    AW_awar *awar_editfield;      // only exists for scanners providing an edit-field. contains content of selected field.
    AW_awar *awar_edit_enabled;   // only (may) exist if awar_editfield exists and scanner provides an enable-edit-toggle. contains value of toggle.
    AW_awar *awar_selected_field; // contains key of selected field (hkey is allowed!)
    AW_awar *awar_mapped_item_ID; // contains id (e.g. name) of currently mapped item (or '<none selected>')
    AW_awar *awar_mark;           // only exists for scanners providing a marked-toggle. contains current status of mark-flag of the currently mapped item.

    AW_root *get_root() const { return aws->get_root(); }

    GBDATA *search_selected_field() const {
        if (gb_item) {
            const char *field_name = awar_selected_field->read_char_pntr();
            if (field_name[0]) {
                return GB_search(gb_item, field_name, GB_FIND);
            }
        }
        return NULp;
    }

    void scan_fields_recursive(GBDATA *gbd, const int depth);
    void scan_list();

    // GUI methods:
    void create_awars(const char *scanner_id, bool have_edit_field, bool have_edit_toggle, bool have_mark_toggle);
    void create_field_selection_list() {
        field_sel = aws->create_selection_list(awar_selected_field->awar_name, 20, 10, true);
    }
    void create_mark_toggle() {
        aws->create_toggle(awar_mark->awar_name);
        awar_mark->add_callback(makeRootCallback(DbScanner::toggle_marked_cb, this));
    }
    void create_field_edit_widgets(const char *edit_pos_fig, const char *edit_enable_pos_fig);

    // callback-handlers:
    void changed_cb(GB_CB_TYPE cbtype);
    void field_changed_cb(GB_CB_TYPE cbtype);
    void keydata_modified_cb(GB_CB_TYPE cbtype);
    void editfield_value_changed_cb();
    void toggle_marked_cb();
    void remap_edit_box();

    // callback-wrappers (called by DB-callback, RootCallback or WindowCallback):
    static void changed_cb(UNFIXED, DbScanner *scanner, GB_CB_TYPE cbtype) { scanner->changed_cb(cbtype); }
    static void field_changed_cb(GBDATA *, DbScanner *scanner, GB_CB_TYPE cbtype) { scanner->field_changed_cb(cbtype); }
    static void keydata_modified_cb(GBDATA*, DbScanner *scanner, GB_CB_TYPE cbtype) { scanner->keydata_modified_cb(cbtype); }
    static void editfield_value_changed_cb(UNFIXED, DbScanner *scanner) { scanner->editfield_value_changed_cb(); }
    static void toggle_marked_cb(UNFIXED, DbScanner *scanner) { scanner->toggle_marked_cb(); }
    static void remap_edit_box(UNFIXED, DbScanner *scanner) { scanner->remap_edit_box(); }

    DbScanner(DB_SCANNERMODE smode, ItemSelector& selector_, AW_window *aws_, GBDATA *gb_main_) :
        selector(selector_),
        aws(aws_),
        gb_main(gb_main_),
        field_sel(NULp),
        gb_item(NULp),
        gb_field(NULp),
        scannermode(smode),
        ignore_editfield_change(false),
        ignore_marktoggle_change(false),
        awar_editfield(NULp),
        awar_edit_enabled(NULp),
        awar_selected_field(NULp),
        awar_mapped_item_ID(NULp),
        awar_mark(NULp)
    {
        arb_assert(&selector);
    }

public:
    // scanner factory:
    static DbScanner *create(GBDATA         *gb_main,
                             const char     *scanner_id,             // unique id (will crash if not)
                             AW_window      *aws,
                             const char     *box_pos_fig,            // xfig-position of field-selection
                             const char     *edit_pos_fig,           // xfig-position of edit-field (or NULp to disable it)
                             const char     *edit_enable_pos_fig,    // xfig-position of edit-toggle (or NULp)
                             DB_SCANNERMODE  scannermode,
                             const char     *mark_pos_fig,           // xfig-position of mark-toggle (or NULp)
                             ItemSelector&   selector);

    const ItemSelector& get_selector() const { return selector; }
    GBDATA *get_gb_main() const { return gb_main; }

    char *get_mapped_item_id() const {
        char *id = NULp;
        if (gb_item) {
            GB_transaction ta(gb_main);
            id = selector.generate_item_id(gb_main, gb_item);
        }
        return id;
    }
    const char *get_mapped_itemID_awarname() const {
        return awar_mapped_item_ID->awar_name;
    }
    void Map(GBDATA *gb_new_item, const char *key_path);

    void RemapToDatabase(GBDATA *gb_new_main) {
        gb_main = gb_new_main;
    }
};

#else
#error db_scanner.hxx included twice
#endif // DB_SCANNER_HXX

